@@ -1,9 +1,12 @@
 
 import crocodile.toolbox as tb
+from crocodile.matplotlib_management import ImShow, FigureSave
 import numpy as np
-from abc import ABC, abstractmethod
+import pandas as pd
+from abc import ABC
 import enum
 from tqdm import tqdm
+import copy
 
 
 # %% ========================== DeepLearning Accessories =================================
@@ -23,7 +26,7 @@ class HyperParam(tb.Struct):
     * When doing multiple experiments, one command in console reminds you of settings used in that run (hp.__dict__).
     * Ease of saving settings of experiments! and also replicating it later.
     """
-    subpath = tb.P('metadata/hyper_params')  # location within model directory where this will be saved.
+    subpath = tb.P('metadata/hyper_param')  # location within model directory where this will be saved.
 
     def __init__(self, **kwargs):
         super().__init__(
@@ -46,20 +49,21 @@ class HyperParam(tb.Struct):
         )
         self._configured = False
         self.device_name = None
-        self.save_type = ["data", "whole", "both"][-1]
+        self.save_type = ["data", "obj", "both"][-1]
         self.update(**kwargs)
 
-    def save(self, path=None, data_only=True, r=False, include_code=False, add_suffix=True):
-        self.save_dir.joinpath(self.subpath / 'hparams.txt').create(parent_only=True).write_text(data=str(self))
-        if self.save_type in {"whole", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.pkl"), add_suffix=False)
+    def save(self, **kwargs):
+        self.save_dir.joinpath(self.subpath / 'hparams.txt').create(parents_only=True).write_text(str(self))
+        if self.save_type in {"data", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.dat.pkl"), add_suffix=False, data_only=True, desc="")
+        if self.save_type in {"obj", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.pkl"), add_suffix=False, data_only=False, desc="")
 
     @classmethod
-    def from_saved(cls, path, *args, r=False, scope=None, **kwargs):  return super(HyperParam, cls).from_saved(path=tb.P(path) / cls.subpath / "hparams.HyperParam.dat.pkl")
-    def __repr__(self): return tb.Struct(self.__dict__).print(config=True, return_str=True)
+    def from_saved_data(cls, path, *args, **kwargs): return super(HyperParam, cls).from_saved_data(tb.P(path) / cls.subpath / "hparams.HyperParam.dat.pkl", *args, **kwargs)
+    def __repr__(self, **kwargs): return "HParams Object with specs:\n" + tb.Struct(self.__dict__).print(config=True, return_str=True)
     @property
     def pkg(self): return __import__("tensorflow") if self.pkg_name == "tensorflow" else (__import__("torch") if self.pkg_name == "torch" else ValueError(f"pkg_name must be either `tensorflow` or `torch`"))
     @property
-    def save_dir(self): return tb.P(self.root) / self.name
+    def save_dir(self): return (tb.P(self.root) / self.name).create()
 
     @property
     def device(self):
@@ -75,19 +79,12 @@ class HyperParam(tb.Struct):
             devices = handle.config.experimental.list_physical_devices('CPU')
             devices += handle.config.experimental.list_physical_devices('GPU')
             device_dict = dict(zip(['cpu', 'gpu0', 'gpu1'], devices))
-
-            if self.device_name is Device.auto:
-                chosen_device = Device.gpu0 if len(devices) > 1 else Device.cpu
-            else:
-                chosen_device = self.device_name
-
+            if self.device_name is Device.auto: chosen_device = Device.gpu0 if len(devices) > 1 else Device.cpu
+            else: chosen_device = self.device_name
             device_str = chosen_device.value if 1 > 0 else "haha"
             if device_str not in device_dict.keys():
-                for i in range(10):
-                    print(f"This machine has no such a device to be chosen! ({device_str})")
-                # Revert to cpu, keep going, instead of throwing an error.
-                device_str = "cpu"
-
+                print(f"This machine has no such a device to be chosen! ({device_str})\n" * 10)
+                device_str = "cpu"  # Revert to cpu, keep going, instead of throwing an error.
             try:
                 device = device_dict[device_str]
                 return device
@@ -98,16 +95,14 @@ class HyperParam(tb.Struct):
 
         elif handle.__name__ == 'torch':
             device = self.device_name
-            if device is Device.auto:
-                return handle.device('cuda:0') if handle.cuda.is_available() else handle.device('cpu')
+            if device is Device.auto: return handle.device('cuda:0') if handle.cuda.is_available() else handle.device('cpu')
+            elif device is Device.cpu: return handle.device('cpu')
             elif device is Device.gpu0:
                 assert handle.cuda.device_count() > 0, f"GPU {device} not available"
                 return handle.device('cuda:0')
             elif device is Device.gpu1:
                 assert handle.cuda.device_count() > 1, f"GPU {device} not available"
                 return handle.device('cuda:1')
-            elif device is Device.cpu:
-                return handle.device('cpu')
             # How to run Torch model on 2 GPUs ?
         else: raise NotImplementedError(f"I don't know how to configure devices for this package {handle}")
 
@@ -133,14 +128,12 @@ class HyperParam(tb.Struct):
                 # logical_gpus = handle.config.experimental.list_logical_devices('GPU')
                 # now, logical gpu is created only for visible device
                 # print(len(devices), "Physical devices,", len(logical_gpus), "Logical GPU")
-
         except AssertionError as e:
             print(e)
             print(f"Trying again with auto-device {Device.auto}")
             self.device_name = Device.auto
             self.config_device()
         except ValueError: print("Cannot set memory growth on non-GPU devices")
-
         except RuntimeError as e:
             print(e)
             print(f"Device already configured, skipping ... ")
@@ -165,27 +158,17 @@ class DataReader(tb.Base):
         self.scaler = None
 
     def save(self, path=None, *args, **kwargs):
-        base = (tb.P(path) if path is not None else self.hp.save_dir).joinpath(self.subpath)
-        if self.hp.save_type in {"whole", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.pkl", add_suffix=False)
+        base = (tb.P(path) if path is not None else self.hp.save_dir).joinpath(self.subpath).create()
+        if self.hp.save_type in {"data", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.dat.pkl", add_suffix=False, data_only=True)
+        if self.hp.save_type in {"obj", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.pkl", add_suffix=False, data_only=False)
 
     @classmethod
-    def from_saved(cls, path, *args, **kwargs):
-        instance = cls(*args, **kwargs)
-        data = (tb.P(path) / cls.subpath / "data_reader.DataReader.dat.pkl").readit()
-        instance.__setstate__(data)
-        return instance
-
+    def from_saved_data(cls, path, *args, **kwargs): return super(DataReader, cls).from_saved_data(tb.P(path) / cls.subpath / "data_reader.DataReader.dat.pkl", *args, **kwargs)
     def __getstate__(self): return dict(specs=self.specs, scaler=self.scaler)
     def __setstate__(self, state): return self.__dict__.update(state)
-    def __repr__(self): return f"DataReader Object with these keys: \n" + tb.Struct(self.__dict__).print(config=True, return_str=True)
+    def __repr__(self): return f"DataReader Object with these keys: \n" + tb.Struct(self.__dict__).print(config=False, return_str=True)
 
     def split_the_data(self, *args, strings=None, **kwargs):
-        """
-        :param args: whatever to be sent to train_test_split
-        :param kwargs: whatever to be sent to train_test_split
-        :param strings:
-        :return:
-        """
         from sklearn.model_selection import train_test_split
         result = train_test_split(*args, test_size=self.hp.test_split, shuffle=self.hp.shuffle, random_state=self.hp.seed, **kwargs)
         self.split = tb.Struct(train_loader=None, test_loader=None)
@@ -197,34 +180,27 @@ class DataReader(tb.Base):
         print(f"================== Training Data Split ===========================")
         self.split.print()
 
-    def sample_dataset(self, aslice=None, dataset="test"):
-        if aslice is None: aslice = slice(0, self.hp.batch_size)
-        keys = self.split.keys().filter(f"'_{dataset}' in x")
-        return tuple([self.split[key][aslice] for key in keys])
+    def sample_dataset(self, aslice=None, indices=None, use_default_slice=False, dataset="test"):
+        keys = self.split.keys().filter(lambda x: f'_{dataset}' in x)
+        selection = indices or aslice
+        res1 = selection or np.random.choice(len(self.split[keys[0]]), size=self.hp.batch_size, replace=False)
+        res2 = selection or slice(0, self.hp.batch_size)
+        selection = res1 if use_default_slice is False and (indices or aslice) is None else (res1 if indices is not None else res2)
+        return tuple([tmp.iloc[selection] if type(tmp := self.split[key]) is pd.DataFrame else tmp[selection] for key in keys])
 
     def get_random_input_output(self, ip_shape=None, op_shape=None):
-        if ip_shape is None:
-            ip_shape = self.specs.ip_shape
-        if op_shape is None:
-            op_shape = self.specs.op_shape
-        if hasattr(self.hp, "precision"):
-            dtype = self.hp.precision
-        else:
-            dtype = "float32"
+        if ip_shape is None: ip_shape = self.specs.ip_shape
+        if op_shape is None: op_shape = self.specs.op_shape
+        dtype = self.hp.precision if hasattr(self.hp, "precision") else "float32"
         ip = np.random.randn(*((self.hp.batch_size,) + ip_shape)).astype(dtype)
         op = np.random.randn(*((self.hp.batch_size,) + op_shape)).astype(dtype)
         return ip, op
 
-    def preprocess(self, *args, **kwargs):
-        _ = args, kwargs, self
-        return args[0]  # acts like identity.
-
-    def postprocess(self, *args, **kwargs):
-        _ = args, kwargs, self
-        return args[0]  # acts like identity
+    def preprocess(self, *args, **kwargs): _ = args, kwargs, self; return args[0]  # acts like identity.
+    def postprocess(self, *args, **kwargs): _ = args, kwargs, self; return args[0]  # acts like identity
 
     def standardize(self):
-        assert self.split is not None, "Load up the data first."
+        assert self.split is not None, "Load up the data first before you standardize it."
         from sklearn.preprocessing import StandardScaler
         self.scaler = StandardScaler()
         self.split.x_train = self.scaler.fit_transform(self.split.x_train)
@@ -234,16 +210,11 @@ class DataReader(tb.Base):
         """
         Assumes numpy inputs
         """
-        if gt is None:
-            labels = None
-            self.plotter = tb.ImShow(pred, labels=labels, sup_titles=names, origin='lower', **kwargs)
-        else:
-            labels = ['Reconstruction', 'Ground Truth']
-            self.plotter = tb.ImShow(pred, gt, labels=labels, sup_titles=names, origin='lower', **kwargs)
+        if gt is None: self.plotter = ImShow(pred, labels=None, sup_titles=names, origin='lower', **kwargs)
+        else: self.plotter = ImShow(pred, gt, labels=['Reconstruction', 'Ground Truth'], sup_titles=names, origin='lower', **kwargs)
 
     def viz(self, *args, **kwargs):
-        """Implement here how you would visualize a batch of input and ouput pair.
-        Assume Numpy arguments rather than tensors."""
+        """Implement here how you would visualize a batch of input and ouput pair. Assume Numpy arguments rather than tensors."""
         _ = self, args, kwargs
         return None
 
@@ -257,15 +228,17 @@ class BaseModel(ABC):
     * :func:`BaseModel.predict` expects a processed input, uese infer and does postprocessing.
     * :func:`BaseModel.predict_from_s` reads, preprocess, then uses predict method.
     * :func:`BseModel.evaluate` Expects processed input and internally calls infer and postprocess methods.
+
+    Functionally or Sequentually built models are much more powerful than Subclassed models. They are faster, have more features, can be plotted, serialized, correspond to computational graphs etc.
     """
-    @abstractmethod
+    # @abstractmethod
     def __init__(self, hp=None, data=None, model=None, compiler=None, history=None):
         self.hp = hp  # should be populated upon instantiation.
         self.model = model  # should be populated upon instantiation.
         self.data = data  # should be populated upon instantiation.
         self.compiler = compiler  # Struct with .losses, .metrics and .optimizer.
         self.history = tb.List() if history is None else history  # should be populated in fit method, or loaded up.
-        self.plotter = tb.SaveType.NullAuto
+        self.plotter = FigureSave.NullAuto
         self.kwargs = None
         self.tmp = None
 
@@ -282,32 +255,26 @@ class BaseModel(ABC):
         if self.hp.pkg_name == 'tensorflow':
             if loss is None: loss = pkg.keras.losses.MeanSquaredError()
             if optimizer is None: optimizer = pkg.keras.optimizers.Adam(self.hp.learning_rate)
-            if metrics is None: metrics = tb.List()  # [pkg.keras.metrics.MeanSquaredError()]
+            if metrics is None: metrics = []  # [pkg.keras.metrics.MeanSquaredError()]
         elif self.hp.pkg_name == 'torch':
             if loss is None: loss = pkg.nn.MSELoss()
             if optimizer is None: optimizer = pkg.optim.Adam(self.model.parameters(), lr=self.hp.learning_rate)
-            if metrics is None: metrics = tb.List()  # [tmp.MeanSquareError()]
+            if metrics is None: metrics = []  # [tmp.MeanSquareError()]
         # Create a new compiler object
-        self.compiler = tb.Struct(loss=loss, optimizer=optimizer, metrics=tb.L(metrics), **kwargs)
+        self.compiler = tb.Struct(loss=loss, optimizer=optimizer, metrics=metrics, **kwargs)
         # in both cases: pass the specs to the compiler if we have TF framework
         if self.hp.pkg.__name__ == "tensorflow" and compile_model: self.model.compile(**self.compiler.__dict__)
 
     def fit(self, viz=False, **kwargs):
-        default_settings = tb.Struct(x=self.data.split.x_train, y=self.data.split.y_train,
-                                     validation_data=(self.data.split.x_test, self.data.split.y_test),
-                                     batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1,
-                                     shuffle=self.hp.shuffle, callbacks=[])
+        default_settings = tb.Struct(x=self.data.split.x_train, y=self.data.split.y_train, validation_data=(self.data.split.x_test, self.data.split.y_test),
+                                     batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1, shuffle=self.hp.shuffle, callbacks=[])
         default_settings.update(kwargs)
         hist = self.model.fit(**default_settings.dict)
-        self.history.append(tb.Struct(tb.copy.deepcopy(hist.history)))
+        self.history.append(tb.Struct(copy.deepcopy(hist.history)))
         # it is paramount to copy, cause source can change.
         if viz: self.plot_loss()
         return self
 
-    def plot_loss(self):
-        total_hist = tb.Struct.concat_values(*self.history)
-        total_hist.plot()
-
     def switch_to_sgd(self, epochs=10):
         print(f'Switching the optimizer to SGD. Loss is fixed to {self.compiler.loss}'.center(100, '*'))
         if self.hp.pkg.__name__ == 'tensorflow': new_optimizer = self.hp.pkg.keras.optimizers.SGD(lr=self.hp.learning_rate * 0.5)
@@ -333,6 +300,16 @@ class BaseModel(ABC):
 
     def postprocess(self, *args, **kwargs): return self.data.postprocess(*args, **kwargs)
     def __call__(self, *args, **kwargs): return self.model(*args, **kwargs)
+    def predict(self, x, **kwargs): return self.postprocess(self.infer(x), **kwargs)
+    def viz(self, *args, **kwargs): self.data.viz(*args, **kwargs)
+    def save_model(self, directory): self.model.save(directory)  # In TF: send only path dir. Save path is saved_model.pb
+    def save_weights(self, directory): self.model.save_weights(directory.joinpath(self.model.name))  # TF: last part of path is file path.
+    @staticmethod
+    def load_model(directory): __import__("tensorflow").keras.models.load_model(directory)  # path to directory. file saved_model.pb is read auto.
+    def load_weights(self, directory): self.model.load_weights(directory.glob('*.data*').__next__().__str__().split('.data')[0])  # requires path to file path.
+    def summary(self): return self.model.summary()
+    def config(self): [print(layer.get_config(), "\n==============================") for layer in self.model.layers]; return None
+    def plot_loss(self): return tb.Struct.concat_values(*self.history).plot()
 
     def infer(self, x):
         """
@@ -343,40 +320,22 @@ class BaseModel(ABC):
         """
         return self.model.predict(x)  # Keras automatically handles special layers.
 
-    def predict(self, x, **kwargs): return self.postprocess(self.infer(x), **kwargs)
-
     def deduce(self, obj, viz=True, **kwargs):
         """Assumes that contents of the object are in the form of a batch."""
         preprocessed = self.preprocess(obj, **kwargs)
         prediction = self.infer(preprocessed)
         postprocessed = self.postprocess(prediction, **kwargs)
         result = tb.Struct(input=obj, preprocessed=preprocessed, prediction=prediction, postprocessed=postprocessed)
-        if viz:
-            self.viz(postprocessed, **kwargs)
+        if viz: self.viz(postprocessed, **kwargs)
         return result
 
-    def viz(self, *args, **kwargs): self.data.viz(*args, **kwargs)
-
     def evaluate(self, x_test=None, y_test=None, names_test=None, idx=None, viz=True, sample=5, **kwargs):
         # ================= Data Procurement ===================================
-        x_test = x_test if x_test is not None else self.data.split.x_test
-        y_test = y_test if y_test is not None else self.data.split.y_test
-        this = self.data.split.names_test if hasattr(self.data.split, "names_test") else range(len(x_test))
-        names_test = names_test if names_test is not None else this
-        if idx is None:
-            def get_rand(x, y):
-                idx_ = np.random.choice(len(x) - sample)
-                return x[idx_:idx_ + sample], y[idx_:idx_ + sample], \
-                    names_test[idx_: idx_ + sample], np.arange(idx_, idx_ + sample)
-            assert self.data is not None, 'Data attribute is not defined'
-            x_test, y_test, names_test, idx = get_rand(x_test, y_test)  # already processed S's
-        else:
-            if type(idx) is int:
-                assert idx < len(x_test), f"Index passed {idx} exceeds length of x_test {len(x_test)}"
-                x_test, y_test, names_test = x_test[idx: idx + 1], y_test[idx: idx + 1], names_test[idx: idx + 1]
-            else: x_test, y_test, names_test = x_test[idx], y_test[idx], names_test[idx]
+        x_test, y_test = x_test if x_test is not None else self.data.split.x_test, y_test if y_test is not None else self.data.split.y_test
+        names_test = names_test if names_test is not None else (self.data.split.names_test if hasattr(self.data.split, "names_test") else range(len(x_test)))
+        idx = np.random.choice(len(x_test) - 1, size=sample, replace=False) if idx is None else (slice(idx, idx + 1, 1) if isinstance(idx, int) else idx)
+        x_test, y_test, names_test = x_test[idx], y_test[idx], names_test[idx]
         # ==========================================================================
-
         prediction = self.infer(x_test)
         loss_dict = self.get_metrics_evaluations(prediction, y_test)
         if loss_dict is not None: loss_dict['names'] = names_test
@@ -400,121 +359,79 @@ class BaseModel(ABC):
             else: name = "unknown"
             # try:  # EAFP principle.
             #     path = a_metric.path  # works for subclasses Metrics
-            # except AttributeError:
-            #
-            #     path = a_metric.__name__  # works for functions.
+            # except AttributeError: path = a_metric.__name__  # works for functions.
             loss_dict[name] = []
-
             for a_prediction, a_y_test in zip(prediction, groun_truth):
-                if hasattr(a_metric, "reset_states"):
-                    a_metric.reset_states()
+                if hasattr(a_metric, "reset_states"): a_metric.reset_states()
                 loss = a_metric(a_prediction[None], a_y_test[None])
                 loss_dict[name].append(np.array(loss).item())
-        return tb.pd.DataFrame(loss_dict)
-
-    def save_model(self, directory): self.model.save(directory)  # In TF: send only path dir. Save path is saved_model.pb
-    def save_weights(self, directory): self.model.save_weights(directory.joinpath(self.model.name))  # TF: last part of path is file path.
-
-    @staticmethod
-    def load_model(directory):
-        import tensorflow as tf
-        return tf.keras.models.load_model(directory)  # path to path. file saved_model.pb is read auto.
-
-    def load_weights(self, directory):
-        name = directory.glob('*.data*').__next__().__str__().split('.data')[0]
-        self.model.load_weights(name)  # requires path to file path.
+        return pd.DataFrame(loss_dict)
 
     def save_class(self, weights_only=True, version='0', **kwargs):
         """Simply saves everything:
-
         1. Hparams
         2. Data specs
         3. Model architecture or weights depending on the following argument.
-
         :param version: Model version, up to the user.
         :param weights_only: self-explanatory
         :return:
-
         """
         self.hp.save()  # goes into the meta path.
         self.data.save()  # goes into the meta path.
-        tb.Save.pickle(obj=self.history, path=self.hp.save_dir / 'metadata/history.pkl')  # goes into the meta path.
-        # model save goes into data path.
-        save_dir = self.hp.save_dir.joinpath(f'{"weights" if weights_only else "model"}_save_v{version}').create()
+        tb.Save.pickle(obj=self.history, path=self.hp.save_dir / 'metadata/history.pkl', verbose=True, desc="Training History")  # goes into the meta path.
+        tb.Experimental.generate_readme(self.hp.save_dir, obj=self.__class__, **kwargs)
+        save_dir = self.hp.save_dir.joinpath(f'{"weights" if weights_only else "model"}_save_v{version}').create()  # model save goes into data path.
         if weights_only: self.save_weights(save_dir)
         else: self.save_model(save_dir)
-        # Saving wrapper_class and model architecture in the main path:
-        tb.Experimental.generate_readme(self.hp.save_dir, obj=self.__class__, **kwargs)
-        print(f'Model class saved successfully!, check out: {self.hp.save_dir.as_uri()}')
+        print(f'SAVED Model Class @ {self.hp.save_dir.as_uri()}')
 
     @classmethod
-    def from_class_weights(cls, path, hparam_class=None, data_class=None, device_name=None):
+    def from_class_weights(cls, path, hparam_class=None, data_class=None, device_name=None, verbose=True):
         path = tb.P(path)
-
-        if hparam_class is not None: hp_obj = hparam_class.from_saved(path)
+        if hparam_class is not None: hp_obj = hparam_class.from_saved_data(path)
         else: hp_obj = (path / HyperParam.subpath + ".HyperParam.pkl").readit()
         if device_name: hp_obj.device_name = device_name
-
-        if data_class is not None: d_obj = data_class.from_saved(path, hp=hp_obj)
+        if data_class is not None: d_obj = data_class.from_saved_data(path, hp=hp_obj)
         else: d_obj = (path / DataReader.subpath / "data_reader.DataReader.pkl").readit()
         d_obj.hp = hp_obj
-
         model_obj = cls(hp_obj, d_obj)
         model_obj.load_weights(path.search('*_save_*')[0])
         model_obj.history = (path / "metadata/history.pkl").readit(notfound=tb.L())
-
-        print(f"Class {model_obj.__class__} Loaded Successfully.")
+        print(f"LOADED {model_obj.__class__}: {model_obj.hp.name}") if verbose else None
         return model_obj
 
     @classmethod
     def from_class_model(cls, path):
         path = tb.P(path)
-        data_obj = DataReader.from_saved(path)
-        hp_obj = HyperParam.from_saved(path)
+        data_obj = DataReader.from_saved_data(path)
+        hp_obj = HyperParam.from_saved_data(path)
         model_obj = cls.load_model(path.search('*_save_*')[0])  # static method.
         wrapper_class = cls(hp_obj, data_obj, model_obj)
         return wrapper_class
 
-    def summary(self): return self.model.summary()
-
-    def config(self):
-        for layer in self.model.layers: print(layer.get_config(), "\n==============================")
-
-    def plot_model(self, **kwargs):
-        """
-        .. note:: Functionally or Sequentually built models are much more powerful than Subclassed models.
-            They are faster, have more features, can be plotted, serialized, correspond to computational graphs etc.
-            Alternative visualization is via tf2onnx then Netron.
-        """
+    def plot_model(self, **kwargs):  # alternative viz via tf2onnx then Netron.
         import tensorflow as tf
-        tf.keras.utils.plot_model(self.model, to_file=self.hp.save_dir / 'model_plot.png',
-                                  show_shapes=True, show_layer_names=True, show_dtype=True,
-                                  expand_nested=True,
-                                  dpi=150, **kwargs)
-        print('Successfully plotted the model, check out \n', (self.hp.save_dir / 'model_plot.png').as_uri())
+        tf.keras.utils.plot_model(self.model, to_file=self.hp.save_dir / 'model_plot.png', show_shapes=True, show_layer_names=True, show_dtype=True, expand_nested=True, dpi=150, **kwargs)
+        print(f"Successfully plotted the model @ {(self.hp.save_dir / 'model_plot.png').as_uri()}")
 
     def build(self, ip_shape=None, verbose=True):
         """ Building has two main uses.
-
-        * Useful to baptize the model, especially when its layers are built lazily. Although this will eventually
-          happen as the first batch goes in. This is a must before showing the summary of the model.
+        * Useful to baptize the model, especially when its layers are built lazily. Although this will eventually happen as the first batch goes in. This is a must before showing the summary of the model.
         * Doing sanity check about shapes when designing model.
         * Sanity check about values and ranges when random normal input is fed.
-
         :param ip_shape:
         :param verbose:
         :return:
         """
         ip, _ = self.data.get_random_input_output(ip_shape=ip_shape)
-        op = self.model(ip)
-        self.tmp = op
+        self.tmp = self.model(ip)  # op
         if verbose:
-            print("============  Build Test ==============")
+            print("Build Test".center(50, '-'))
             print(f"Input shape = {ip.shape}")
-            print(f"Output shape = {op.shape}")
+            print(f"Output shape = {self.tmp.shape}")
             print("Stats on output data for random normal input:")
-            print(tb.pd.DataFrame(np.array(op).flatten()).describe())
-            print("----------------------------------------", '\n\n')
+            print(pd.DataFrame(np.array(self.tmp).flatten()).describe())
+            print("-" * 100 + '\n' * 2)
 
 
 class Ensemble(tb.Base):
@@ -541,7 +458,7 @@ class Ensemble(tb.Base):
             for i in tqdm(range(size)):
                 hp = self.hp_class()
                 hp.name = str(hp.name) + f'__model__{i}'
-                datacopy = tb.copy.copy(self.data)  # shallow copy
+                datacopy = copy.copy(self.data)  # shallow copy
                 datacopy.hp = hp
                 self.models.append(model_class(hp, datacopy))
         self.performance = None
@@ -562,8 +479,7 @@ class Ensemble(tb.Base):
         self.performance = tb.L()
         for i in range(self.size):
             print('\n\n', f" Training Model {i} ".center(100, "*"), '\n\n')
-            if shuffle_train_test:
-                self.data.split_my_data_now(seed=np.random.randint(0, 1000))  # shuffle data (shared among models)
+            if shuffle_train_test: self.data.split_my_data_now(seed=np.random.randint(0, 1000))  # shuffle data (shared among models)
             self.models[i].fit(**kwargs)
             self.performance.append(self.models[i].evaluate(idx=slice(0, -1), viz=False))
             if save:
@@ -571,9 +487,7 @@ class Ensemble(tb.Base):
                 self.performance.save_pickle(self.hp_class.save_dir / "performance.pkl")
         print("\n\n", f" Finished fitting the ensemble ".center(100, ">"), "\n")
 
-    def clear_memory(self):
-        # t.cuda.empty_cache()
-        pass
+    def clear_memory(self): pass  # t.cuda.empty_cache()
 
 
 class Losses:
@@ -658,9 +572,7 @@ class HPTuning:
                 accuracy = self.run(param_dict)
                 self.pkg.summary.scalar(self.acc_metric, accuracy, step=1)
 
-    def optimize(self):
-        self.gen_writer()
-        self.loop()
+    def optimize(self): self.gen_writer(); self.loop()
 
 
 class KerasOptimizer:
